#if !defined(_CIW_GAME_ANIMATION_H_)
#define _CIW_GAME_ANIMATION_H_

#include "IwGeom.h"
#include "IwList.h"
#include "IwHashString.h"
#include "IwGameUtil.h"
#include "IwGameXoml.h"

class CIwGameAnimInstance;

//
//
//
//
// Animation frame data types
//
//
//
//
struct CIwGameAnimFrame
{
	enum eIwGameAnimType
	{
		FT_Invalid,
		FT_Bool,
		FT_Float,
		FT_Vec2,
		FT_Vec3,
		FT_Vec4,
		FT_Rect,
		FT_String,
		FT_Custom,
	};

	float					Time;			// Time at which this frame should be active

	CIwGameAnimFrame() {}
	virtual ~CIwGameAnimFrame() {}
};

struct CIwGameAnimFrameBool: public CIwGameAnimFrame
{
	bool	data;
};

struct CIwGameAnimFrameFloat: public CIwGameAnimFrame
{
	float	data;
};

struct CIwGameAnimFrameVec2: public CIwGameAnimFrame
{
	CIwFVec2 data;
};

struct CIwGameAnimFrameVec3: public CIwGameAnimFrame
{
	CIwFVec3 data;
};

struct CIwGameAnimFrameVec4: public CIwGameAnimFrame
{
	CIwFVec4 data;
};

struct CIwGameAnimFrameRect: public CIwGameAnimFrame
{
	CIwRect data;
};

struct CIwGameAnimFrameString: public CIwGameAnimFrame
{
	CIwGameString data;
};

//
//
//
//
// IIwGameAnimTarget - An animation target object
//
// 
// 
//
class IIwGameAnimTarget
{
public:
	virtual bool	UpdateFromAnimation(CIwGameAnimInstance *animation) = 0;
};

//
//
//
//
// CIwGameAnim - A collection of animation frames
//
//
// 
//
class CIwGameAnim : public IIwGameXomlResource
{
public:
	enum eiwGameAnimStatus
	{
		AS_Stopped, 
		AS_Playing, 
		AS_Paused, 
	};
	typedef CIwList<CIwGameAnimFrame*>::iterator _Iterator;
	_Iterator begin()					{ return Frames.begin(); }
	_Iterator end()						{ return Frames.end(); }

protected:
	// Properties
#if defined(_DEBUG)
	CIwGameString						Name;						// Name of animation
#endif
	unsigned int						NameHash;					// Name of animation (stored as an hash for speed)
	CIwGameAnimFrame::eIwGameAnimType	Type;						// Type of animation
	float								Duration;					// Duration of animation
	CIwList<CIwGameAnimFrame*>			Frames;						// Animation frames

public:
	void								setName(const char* name)
	{
#if defined(_DEBUG)
		Name = name;
#endif
		NameHash = IwHashString(name);
	}
#if defined(_DEBUG)
	CIwGameString&						getName(const char* name)						{ return Name; }
#endif
	unsigned int						getNameHash()									{ return NameHash; }
	void								setType(CIwGameAnimFrame::eIwGameAnimType type)	{ Type = type; }
	CIwGameAnimFrame::eIwGameAnimType	getType() const									{ return Type; }
	float								getDuration() const								{ return Duration; }
	void								setDuration(float duration)						{ Duration = duration; }
	void								addFrame(CIwGameAnimFrame* frame);
	bool								getCurrentFrame(CIwGameAnimFrame*& current, CIwGameAnimFrame*& next, float current_time);
	void								CalculateCurrentData(float current_time, CIwGameAnimFrame* current_data);
	int									getFrameCount()									{ return Frames.size(); }
	// Properties end

private:
	void								CalculateCurrentData();

public:
	CIwGameAnim() : IIwGameXomlResource(), Duration(0) { setClassType("Animation"); }
	virtual ~CIwGameAnim()
	{
		for (_Iterator it = Frames.begin(); it != Frames.end(); it++)
		{
			delete (*it);
		}
		Frames.clear();
	}

	void								GenerateAtlasFrames(int count, int frame_w, int frame_h, int start_x, int start_y, int pitch_x, int pitch_y, int image_width, float frame_speed);

	// Implementation of IIwGameXomlClass interface
	bool					LoadFromXoml(IIwGameXomlResource* parebt, bool load_children, CIwGameXmlNode* node);
};

//
//  CIwGameAnimCreator - Creates an instance of an animation object
//
class CIwGameAnimCreator : public IIwGameXomlClassCreator
{
public:
	CIwGameAnimCreator()
	{
		setClassName("Animation");
	}
	IIwGameXomlResource*	CreateInstance(IIwGameXomlResource* parent)	{ return new CIwGameAnim(); }
};

//
//
//
//
// CIwGameAnimManager - A collection of animation
//
//
// 
//
class CIwGameAnimManager
{
public:
	typedef CIwList<CIwGameAnim*>::iterator _Iterator;
	_Iterator begin()						{ return Animations.begin(); }
	_Iterator end()							{ return Animations.end(); }

protected:
	// Properties
	CIwList<CIwGameAnim*>	Animations;
public:
	void					addAnimation(CIwGameAnim* anim) { Animations.push_back(anim); }
	void					removeAnimation(CIwGameAnim* anim)
	{
		for (_Iterator it = begin(); it != end(); it++)
		{
			if (*it == anim)
			{
				delete (*it);
				Animations.erase(it);
				break;
			}
		}
	}
	CIwGameAnim*			findAnimation(const char* name, bool global_search = true);
	CIwGameAnim*			findAnimation(unsigned int name_hash, bool global_search = true);
	int						getAnimationCount()					{ return Animations.size(); }
	// Properties end

private:

public:
	CIwGameAnimManager() {	}
	virtual ~CIwGameAnimManager()
	{
		for (_Iterator it = begin(); it != end(); it++)
		{
			delete (*it);
		}
		Animations.clear();
	}
};

//
//
//
//
// CUIAnimationInstance - An instance of an animation object
//
//
// 
//
class CIwGameAnimInstance
{
public:
protected:
	// Properties
	CIwGameAnim*					Animation;				// Animation to use
	bool							Managed;				// True if managed by a CIwGameAnimInstanceManager
	int								RepeatCount;			// Number of times to replay the animation
	float							CurrentTime;			// Current animation time (0 is start of animation)
	CIwGameAnimFrame*				CurrentData;			// The current interpolated data for the current time
	CIwGameAnim::eiwGameAnimStatus	Status;					// Status of the animation
	IIwGameAnimTarget*				Target;					// Target object to update
#if defined(_DEBUG)
	CIwGameString					TargetProperty;			// Target object property name to update
#endif
	unsigned int					TargetPropertyHash;		// Target object property name hash to update
	float							Delay;					// Amount of time to delay starting the animation
	bool							DelayExpired;			// true if initial delay has expired
	CIwGameCallback					StartedCallback;		// Callback which is called when the animation starts playing
	CIwGameCallback					StoppedCallback;		// Callback which is called when the animation stops playing
	CIwGameCallback					LoopedCallback;			// Callback which is called when the animation loops

public:
	CIwGameAnim*		getAnimation()									{ return Animation; }
	void				setAnimation(CIwGameAnim* anim);
	void				setManaged(bool managed)						{ Managed = managed; }
	bool				isManaged() const								{ return Managed; }
	int					getRepeatCount() const							{ return RepeatCount; }
	void				setRepeatCount(int repeat_count)				{ RepeatCount = repeat_count; }
	float				getCurrentTime() const							{ return CurrentTime; }
	bool				isFinished() const								{ if (RepeatCount == 0) return false; else return RepeatedCount >= RepeatCount; }
	CIwGameAnimFrame*	getCurrentData()								{ return CurrentData; }
#if defined(_DEBUG)
	CIwGameString&		getTargetProperty()								{ return TargetProperty; }
#endif
	unsigned int		getTargetPropertyHash()							{ return TargetPropertyHash; }
	IIwGameAnimTarget*	getTargetElement()								{ return Target; }
	void				setTarget(IIwGameAnimTarget* element, const char* property_name)
	{
		Target = element;
#if defined(_DEBUG)
		TargetProperty = property_name;
#endif
		TargetPropertyHash = CIwGameString::CalculateHash(property_name);
	}
	void				setTargetElement(IIwGameAnimTarget* element)	{ Target = element; }
	float				getDelay() const								{ return Delay; }
	void				setDelay(float delay)							{ Delay = delay; }
	bool				isPlaying() const								{ return Status == CIwGameAnim::AS_Playing; }
	bool				isPaused() const								{ return Status == CIwGameAnim::AS_Paused; }
	bool				isDelayed() const								{ return !DelayExpired; }
	void				restart()										{ Status = CIwGameAnim::AS_Playing; CurrentTime = 0; DelayExpired = false; RepeatedCount = 0; }
	void				play()											{ Status = CIwGameAnim::AS_Playing; }
	void				stop()											{ Status = CIwGameAnim::AS_Stopped; CurrentTime = 0; DelayExpired = false; RepeatedCount = 0; }
	void				pause()											{ Status = CIwGameAnim::AS_Paused; }
	void				setStartedCallback(CIwGameCallback callback)	{ StartedCallback = callback; }
	void				setStoppedCallback(CIwGameCallback callback)	{ StoppedCallback = callback; }
	void				setLoopedCallback(CIwGameCallback callback)		{ LoopedCallback = callback; }
	// Properties end

private:
	int					RepeatedCount;

public:
	CIwGameAnimInstance() : Managed(false), Animation(NULL), RepeatCount(0), CurrentTime(0), RepeatedCount(0), Delay(0), DelayExpired(false), Status(CIwGameAnim::AS_Stopped), 
							StartedCallback(NULL), StoppedCallback(NULL), LoopedCallback(NULL) { }
	virtual ~CIwGameAnimInstance()
	{
		delete CurrentData;
	}

	virtual void	Update(float dt);
};


//
//
//
//
// CIwGameAnimTimeline - An animation timeline
//
//
// 
//
class CIwGameAnimTimeline : public IIwGameXomlResource
{
public:
	typedef CIwList<CIwGameAnimInstance*>::iterator _Iterator;
	_Iterator begin()						{ return Animations.begin(); }
	_Iterator end()							{ return Animations.end(); }

protected:
	// Properties
#if defined(_DEBUG)
	CIwGameString					Name;						// Name of animation
#endif
	unsigned int					NameHash;					// Name of animation (stored as an hash for speed)
	CIwList<CIwGameAnimInstance*>	Animations;
	bool							Managed;					// True if managed by a CIwGameAnimTimelineManager
public:
	void							setName(const char* name)
	{
#if defined(_DEBUG)
		Name = name;
#endif
		NameHash = IwHashString(name);
	}
#if defined(_DEBUG)
	CIwGameString&					getName(const char* name)	{ return Name; }
#endif
	unsigned int					getNameHash()						{ return NameHash; }
	void							addAnimation(CIwGameAnimInstance* anim) { Animations.push_back(anim); }
	void							removeAnimation(CIwGameAnimInstance* anim)
	{
		for (_Iterator it = begin(); it != end(); it++)
		{
			if (*it == anim)
			{
				delete (*it);
				Animations.erase(it);
				break;
			}
		}
	}
	CIwGameAnimInstance*			findAnimation(const char* name);
	CIwGameAnimInstance*			findAnimation(unsigned int name_hash);
	void							play();
	void							stop();
	void							pause();
	void							restart();
	int								getAnimationCount()					{ return Animations.size(); }
	void							setTargetElement(IIwGameAnimTarget* target)
	{
		for (_Iterator it = begin(); it != end(); it++)
			(*it)->setTargetElement(target);
	}
	void							setManaged(bool managed)			{ Managed = managed; }
	bool							isManaged() const					{ return Managed; }
	// Properties end

private:

public:
	CIwGameAnimTimeline() : Managed(false) { setClassType("Timeline"); }
	virtual ~CIwGameAnimTimeline()
	{
		for (_Iterator it = begin(); it != end(); it++)
			delete (*it);
		Animations.clear();
	}

	virtual void	Update(float time_delta)
	{
		for (_Iterator it = begin(); it != end(); it++)
			(*it)->Update(time_delta);
	}

	// Implementation of IIwGameXomlClass interface
	bool					LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node);
};

//
//  CIwGameAnimTimelineCreator - Creates an instance of an animation object
//
class CIwGameAnimTimelineCreator : public IIwGameXomlClassCreator
{
public:
	CIwGameAnimTimelineCreator()
	{
		setClassName("Timeline");
	}
	IIwGameXomlResource*	CreateInstance(IIwGameXomlResource* parent)	{ return new CIwGameAnimTimeline(); }
};

//
//
//
//
// CIwGameAnimTimelineManager - A collection of timelines
//
//
// 
//
class CIwGameAnimTimelineManager
{
public:
	typedef CIwList<CIwGameAnimTimeline*>::iterator _Iterator;
	_Iterator begin()						{ return Timelines.begin(); }
	_Iterator end()							{ return Timelines.end(); }

protected:
	// Properties
	CIwList<CIwGameAnimTimeline*>	Timelines;
public:
	void					addTimeline(CIwGameAnimTimeline* timeline) { Timelines.push_back(timeline); timeline->setManaged(true); }
	void					removeTimeline(CIwGameAnimTimeline* timeline)
	{
		for (_Iterator it = begin(); it != end(); it++)
		{
			if (*it == timeline)
			{
				delete (*it);
				Timelines.erase(it);
				break;
			}
		}
	}
	CIwGameAnimTimeline*	findTimeline(const char* name, bool global_search = true);
	CIwGameAnimTimeline*	findTimeline(unsigned int name_hash, bool global_search = true);
	int						getTimelineCount()					{ return Timelines.size(); }
	// Properties end

private:

public:
	CIwGameAnimTimelineManager() {	}
	virtual ~CIwGameAnimTimelineManager()
	{
		for (_Iterator it = begin(); it != end(); it++)
		{
			delete (*it);
		}
		Timelines.clear();
	}
};


#endif	// _CANIM_H_