// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#if !defined(_CIW_GAME_IMAGE_H_)
#define _CIW_GAME_IMAGE_H_

#include "IwGx.h"
#include "Iw2D.h"
#include "IwList.h"
#include "IwGameUtil.h"
#include "IwGameXoml.h"
#include "IwGameFile.h"

class CIwGameImageManager;

//
//
//
//
// CIwGameImage - Represents a bitmapped image
//
// Note that images can be either loaded on demand (loaded on first call to getImage2D() or pre loaded by calling Load()
//
//
//
//
class CIwGameImage : public IIwGameXomlResource
{
public:
	enum eCIwGameImage_State
	{
		CIwGameImage_State_Invalid, 
		CIwGameImage_State_Loaded, 
	};

	// Properties
protected:
	CIwGameImageManager*	Parent;				// Parent image manager that allocated us
	unsigned int			NameHash;			// Hashed value of resuorce name (for quick seacrhing)
	eCIwGameImage_State		State;				// State of image
	CIwResGroup*			ResourceGroup;		// Resource group that our image lives in
	CIw2DImage*				Image2D;			// Iw2D Image
	int						Width;				// Pixel width of image
	int						Height;				// Pixl height of image
public:
	void					setParent(CIwGameImageManager* parent)	{ Parent = parent; }
	CIwGameImageManager*	getParent()								{ return Parent; }
	void					setName(const char* name)				{ NameHash = IwHashString(name); }
	unsigned int			getNameHash() const						{ return NameHash; }
	eCIwGameImage_State		getState() const						{ return State; }
	CIw2DImage*				getImage2D()							{ if (Load()) return Image2D; return NULL; }
	int						getWidth() const						{ return Width; }
	int						getHeight() const						{ return Height; }
	CIwGameImage*			getCopy();
	// Properties End
protected:
	CIwGameFile*			File;				// File object (if image if file based)
	bool					DecompressJPEG(char* jpeg_data, int jpeg_data_size);

public:
	CIwGameImage() : Parent(NULL), ResourceGroup(NULL), Image2D(NULL), File(NULL), State(CIwGameImage_State_Invalid), Width(0), Height(0)	{ setClassType("Image"); }
	virtual ~CIwGameImage()
	{
		SAFE_DELETE(Image2D)
		SAFE_DELETE(File)
	}

	void			Init(const char* ResourceName, CIwResGroup* resource_group)		// Init an image from an image located within a resource group (image is not loaded)
	{
		NameHash = CIwGameString::CalculateHash(ResourceName);
		ResourceGroup = resource_group;
	}

	bool			Init(void* memory_file, int memory_file_size);	// Init an image from a memory based file (image is loaded)
	void			Init(const char* filename)						// Init an image from a file (image is not loaded)
	{
		File = new CIwGameFile();
		File->setFilename(filename);
	}

	bool			Load(bool blocking = true);						// Force load the image

	// Utility
//	char*			CreateJPEG();
	uint8*			CreatePNG(int& size);

	// Implementation of IIwGameXomlClass interface
	bool			LoadFromXoml(IIwGameXomlResource* parebt, bool load_children, CIwGameXmlNode* node);

	// Internal
	void			FinishLoad();									// Called back when aysnc loading is completed
};

//
//  CIwGameImageCreator - Creates an instance of an image object
//
class CIwGameImageCreator : public IIwGameXomlClassCreator
{
public:
	CIwGameImageCreator()
	{
		setClassName("Image");
	}
	IIwGameXomlResource*	CreateInstance(IIwGameXomlResource* parent)	{ return new CIwGameImage(); }
};

//
//
//
//
// CIwGameImageManager - Manages a collection of bitmapped images
//
//
//
// 
class CIwGameImageManager
{
public:
	// Public access to image iteration
	typedef CIwList<CIwGameImage*>::iterator _Iterator;
	_Iterator				begin() { return Images.begin(); }
	_Iterator				end()	{ return Images.end(); }
	// Properties
protected:
	CIwList<CIwGameImage*>	Images;		// Collection of images
public:
	void					addImage(const char* resource_name, CIwResGroup* resource_group, bool preload = false);	// Creates a new image and adds it to the image manager
	void					addImage(CIwGameImage *image);									// Adds an image to the manager but does not load the image
	void					removeImage(CIwGameImage* image);								// Removes image from the manager and deletes the image
	void					removeImage(unsigned int name_hash);							// Removes image from the manager and deletes the image
	CIwGameImage*			findImage(unsigned int name_hash, bool global_search = true);
	CIwGameImage*			findImage(const char* name, bool global_search = true);
	void					clearImages();													// Removes and deletes all images in the manager
	// Properties End
protected:

public:
	CIwGameImageManager()			{}
	virtual ~CIwGameImageManager()	{ clearImages(); }
};

#endif	// _CIW_GAME_IMAGE_H_
