// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#if !defined(_IW_GAME_XOML_H_)
#define _IW_GAME_XOML_H_


#include "s3efile.h"
#include "IwGameString.h"
#include "IwGameUtil.h"
#include "IwGameXml.h"

class CIwGame;

// 
// 
//
//
// CIwGameXomlNames - pre-hashed strings of classes, attributes and certain attributes data used to speed up string searchimng
// 
// 
//
//
struct CIwGameXomlNames
{
public:
	static unsigned int Game_Hash;
	static unsigned int Scene_Hash;
	static unsigned int Resource_Hash;
	static unsigned int Actor_Hash;
	static unsigned int ActorImage_Hash;
	static unsigned int ResourceGroup_Hash;
	static unsigned int Animation_Hash;
	static unsigned int AnimationCollection_Hash;

	static unsigned int Name_Hash;
	static unsigned int Current_Hash;
	static unsigned int CanvasSize_Hash;
	static unsigned int FixAspect_Hash;
	static unsigned int LockWidth_Hash;
	static unsigned int Extents_Hash;
	static unsigned int AllowSuspend_Hash;
	static unsigned int Clipping_Hash;
	static unsigned int Active_Hash;
	static unsigned int Visible_Hash;
	static unsigned int Layers_Hash;
	static unsigned int Colliders_Hash;
	static unsigned int Type_Hash;
	static unsigned int Position_Hash;
	static unsigned int Velocity_Hash;
	static unsigned int VelocityDamping_Hash;
	static unsigned int Angle_Hash;
	static unsigned int AngularVelocity_Hash;
	static unsigned int AngularVelocityDamping_Hash;
	static unsigned int Scale_Hash;
	static unsigned int Colour_Hash;
	static unsigned int Layer_Hash;
	static unsigned int Collidable_Hash;
	static unsigned int CollisionSize_Hash;
	static unsigned int CollisionRect_Hash;
	static unsigned int WrapPosition_Hash;
	static unsigned int Image_Hash;
	static unsigned int Size_Hash;
	static unsigned int GroupFile_Hash;
	static unsigned int Frame_Hash;
	static unsigned int Value_Hash;
	static unsigned int Time_Hash;
	static unsigned int Repeat_Hash;
	static unsigned int Duration_Hash;
	static unsigned int SrcRect_Hash;
	static unsigned int Location_Hash;
	static unsigned int Preload_Hash;
	static unsigned int Blocking_Hash;
	static unsigned int AutoPlay_Hash;
	static unsigned int Anim_Hash;
	static unsigned int StartAtTime_Hash;
	static unsigned int Target_Hash;
	static unsigned int Timeline_Hash;

	static unsigned int Bool_Hash;
	static unsigned int Float_Hash;
	static unsigned int Vec2_Hash;
	static unsigned int Vec3_Hash;
	static unsigned int Vec4_Hash;
	static unsigned int Rect_Hash;
	static unsigned int String_Hash;

};

// 
// 
//
//
// CIwGameXomlClass - Base class that is used by classes that can be instanced from XONML markup
// 
// To create a class that can be instantiated from XOML you should:
// 1. Derive the class from CIwGameXomlClass
// 2. Set its XOMl base type (Game, Actor, Scene, Animation etc..)
// 3. Implement a class derived from IIwGameXomlClassCreator that implements CreateInstance() which creates an instance of the class
// 4. Add the class type to the list of class creators in IW_GAME_XOML using addClass()
//
//
//
//
class IIwGameXomlResource
{
	// Properties
protected:
	unsigned int			ClassType;				// Base type of XOML class (e.g. Scene, Actor, Animation, Game etc..)
public:
	void					setClassType(const char* name)  { ClassType = CIwGameString::CalculateHash(name); }
	unsigned int			getClassType() const			{ return ClassType; }
	// Properties end
public:
	virtual bool LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node) = 0;
};

class IIwGameXomlClassCreator
{
	// Properties
protected:
#if defined(_DEBUG)
	CIwGameString			ClassName;				// Name of class
#endif
	unsigned int			ClassNameHash;			// Hashed name of class
public:
	void					setClassName(const char* name)
	{
#if defined(_DEBUG)
		ClassName = name;
#endif
		ClassNameHash = CIwGameString::CalculateHash(name);
	}
	unsigned int			getClassNameHash() const							{ return ClassNameHash; }
	// Properties end
public:
	virtual IIwGameXomlResource* CreateInstance(IIwGameXomlResource* parent) = 0;
};

#define IW_GAME_XOML	CIwGameXoml::getInstance()

// 
// 
//
//
// CIwGameXoml - CIwGameXoml is the main controller responsible for instantiating objects from XOML based files
// 
// 
// 
//
class CIwGameXoml
{
public:
	CDEFINE_SINGLETON(CIwGameXoml)

	// Public access to class creator iteration
	typedef CIwList<IIwGameXomlClassCreator*>::iterator _Iterator;
	_Iterator				begin() { return ClassCreators.begin(); }
	_Iterator				end() { return ClassCreators.end(); }

protected:
	// Properties
public:
	void		addClass(IIwGameXomlClassCreator* creator)
	{
		ClassCreators.push_back(creator);
	}
	// Properties end

protected:
	CIwList<IIwGameXomlClassCreator*> ClassCreators;

public:
	void		Init();
	void		Release();

	bool		Process(IIwGameXomlResource* parent, const char* filename);
	bool		Process(IIwGameXomlResource* parent, CIwGameXmlNode* node);

};

#endif	// _IW_GAME_XOML_H_
