// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGameResources.h"
#include "IwGameUtil.h"
#include "IwGameString.h"
#include "IwGameAnim.h"
#include "IwGameImage.h"
#include "IwGameScene.h"


//
//
//
//
// CIwGameResourceGroup implementation
//
//
//
//
CIwResGroup* CIwGameResourceGroup::getResourceGroup()
{
	// If group has not been oaded then load it
	if (!Loaded)
		Load();

	return ResourceGroup;
}

bool CIwGameResourceGroup::Load()
{
	if (Loaded)
		return true;

	ResourceGroup = IwGetResManager()->LoadGroup(GroupFilename.c_str());
	if (ResourceGroup != NULL)
	{
		Loaded = true;
		return true;
	}

	return false;
}

void CIwGameResourceGroup::Destroy()
{
	if (Loaded)
	{
		IwGetResManager()->DestroyGroup(ResourceGroup);
		ResourceGroup = NULL;
		Loaded = false;
	}
}

bool CIwGameResourceGroup::LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node)
{
	// Process Actor attributes
	bool preload = false;
	for (CIwGameXmlNode::_AttribIterator it = node->attribs_begin(); it != node->attribs_end(); it++)
	{
		unsigned int name_hash = (*it)->getName().getHash();

		if (name_hash == CIwGameXomlNames::Name_Hash)
		{
			setGroupName((*it)->GetValue().c_str());
		}
		else
		if (name_hash == CIwGameXomlNames::GroupFile_Hash)
		{
			setGroupFilename((*it)->GetValue().c_str());
		}
		else
		if (name_hash == CIwGameXomlNames::Preload_Hash)
		{
			preload = (*it)->GetValueAsBool();
		}
	}

	// Process inner tags
	if (!IW_GAME_XOML->Process(this, node))
		return false;

	// If we are declared inside a scene then animation is local to the scene
	if (parent != NULL && parent->getClassType() == CIwGameXomlNames::Scene_Hash)
	{
		((CIwGameScene*)parent)->getResourceGroupManager()->addGroup(this);
	}
	else
	{
		IW_GAME_GLOBAL_RESOURCES->getResourceGroupManager()->addGroup(this);
	}

	// Preload the resorce group if requested
	if (preload)
		Load();

	return true;
}

//
//
//
//
// CIwGameResourceGroupCreator implementation
//
//
//
//
IIwGameXomlResource* CIwGameResourceGroupCreator::CreateInstance(IIwGameXomlResource* parent)
{
	return new CIwGameResourceGroup();
}



//
//
//
//
// CIwGameResourceGroupManager implementation
//
//
//
//
void CIwGameResourceGroupManager::addGroup(const char* group_name, const char* group_filename, bool preload)
{
	CIwGameResourceGroup* group = new CIwGameResourceGroup();
	group->setGroupName(group_name);
	group->setGroupFilename(group_filename);
	Groups.push_back(group);

	if (preload)
	{
		group->Load();
	}
}

void CIwGameResourceGroupManager::addGroup(CIwGameResourceGroup *group)
{
	Groups.push_back(group);
}

void CIwGameResourceGroupManager::removeGroup(CIwGameResourceGroup* group)
{
	removeGroup(group->getGroupName().getHash());
}

void CIwGameResourceGroupManager::removeGroup(unsigned int name_hash)
{
	// Renove named image from the manager
	for (_Iterator it = Groups.begin(); it != Groups.end(); ++it)
	{
		if (name_hash == (*it)->getGroupName().getHash())
		{
			delete *it;
			Groups.erase(it);
			break;
		}
	}
}

CIwGameResourceGroup* CIwGameResourceGroupManager::findGroup(unsigned int name_hash, bool global_search)
{
	// Find and eturn named image
	for (_Iterator it = Groups.begin(); it != Groups.end(); ++it)
	{
		if (name_hash == (*it)->getGroupName().getHash())
			return *it;
	}

	if (global_search)
	{
		// Not found so check global resources
		return IW_GAME_GLOBAL_RESOURCES->getResourceGroupManager()->findGroup(name_hash, false);
	}

	return NULL;
}

CIwGameResourceGroup* CIwGameResourceGroupManager::findGroup(const char* name, bool global_search)
{
	return findGroup(CIwGameString::CalculateHash(name), global_search);
}

void CIwGameResourceGroupManager::clearGroups()
{
	// Remove and delete all images from the manager
	for (_Iterator it = Groups.begin(); it != Groups.end(); ++it)
	{
		delete *it;
	}
	Groups.clear();
}

//
//
//
//
// CIwGameGlobalResources implementation
//
//
//
//

CDECLARE_SINGLETON(CIwGameGlobalResources)

void CIwGameGlobalResources::Init()
{
	ImageManager = new CIwGameImageManager();
	ResourceGroupManager = new CIwGameResourceGroupManager();
	AnimManager = new CIwGameAnimManager();
	TimelineManager = new CIwGameAnimTimelineManager();
}

void CIwGameGlobalResources::Release()
{
	SAFE_DELETE(ImageManager)
	SAFE_DELETE(ResourceGroupManager)
	SAFE_DELETE(AnimManager)
	SAFE_DELETE(TimelineManager)
}