// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGame.h"
#include "IwGameScene.h"
#include "IwGameActor.h"
#include "IwGameCamera.h"
#include "IwGameString.h"

//
//
//
//
// CIwGameScene implementation
//
//
//
//
CIwGameScene::~CIwGameScene()
{
	// Only delete an unmanaged timeline
	if (Timeline != NULL)
	{
		if (!Timeline->isManaged())
		{
			SAFE_DELETE(Timeline)
		}
	}

	SAFE_DELETE(Collidables)
	clearActors();
	SAFE_DELETE(SpriteManager)
	SAFE_DELETE(ImageManager)
	SAFE_DELETE(ResourceGroupManager)
	SAFE_DELETE(AnimManager)
	SAFE_DELETE(TimelineManager)
	SAFE_DELETE(Camera)
}

void CIwGameScene::setVirtualTransform(int required_width, int required_height, float angle, bool fix_aspect, bool lock_width)
{
	float scale_x, scale_y;
	int screen_width = IwGxGetScreenWidth();
	int screen_height = IwGxGetScreenHeight();
	
	ScreenSize.x = screen_width;
	ScreenSize.y = screen_height;
	VirtualSize.x = required_width;
	VirtualSize.y = required_height;
	if (fix_aspect)
	{
		// A fixed aspect ensures that the aspect ratio of the scene fits the native screen aspect ratio
		if (lock_width)
			scale_x = scale_y = (float)screen_width / required_width;
		else
			scale_x = scale_y = (float)screen_height / required_height;
	}
	else
	{
		// Simply scale to fit native screen resolution
		scale_x = (float)screen_width / required_width;
		scale_y = (float)screen_height / required_height;
	}

	// Build rotation / scale tranform
	CIwMat2D scale;
	scale.SetIdentity();
	scale.m[0][0] = (iwfixed)(scale_x * IW_GEOM_ONE);
	scale.m[1][1] = (iwfixed)(scale_y * IW_GEOM_ONE);
	VirtualTransform.SetRot((iwangle)((angle * IW_ANGLE_2PI) / 360.0f));
	VirtualTransform.PostMult(scale);

	// Cantre scene in middle of screen
	VirtualTransform.t.x = screen_width  / 2;
	VirtualTransform.t.y = screen_height  / 2;

	// If camera attacjhed then force transform to become dirty so that the scene transform gets rebuilt
	if (Camera != NULL)
	{
		Camera->forceTransformDirty();
	}
	else
	{
		// If no camera attached then just set the scene transform to the virtual transform
		Transform = VirtualTransform;

		// Transfer scene transform to sprite manager
		if (SpriteManager != NULL)
		{
			SpriteManager->setTransform(Transform);
		}
	}
}

int	 CIwGameScene::Init(int max_collidables, int max_layers)
{
	// Allocate collidables list
	Timeline = NULL;
	Collidables = new CIwGameActor*[max_collidables];
	MaxCollidables = max_collidables;
	MaxLayers = max_layers;

	// Allocate managers
	SpriteManager = new CIwGameSpriteManager();
	SpriteManager->Init(max_layers);
	ImageManager = new CIwGameImageManager();
	ResourceGroupManager = new CIwGameResourceGroupManager();
	AnimManager = new CIwGameAnimManager();
	TimelineManager = new CIwGameAnimTimelineManager();

	Colour.r = 0xff;
	Colour.g = 0xff;
	Colour.b = 0xff;
	Colour.a = 0xff;

	return 1;
}

void CIwGameScene::addActor(CIwGameActor *actor)
{
	// Add an actor to the scene
	Actors.push_back(actor);
	actor->setScene(this);
}

void CIwGameScene::removeActor(CIwGameActor* actor)
{
	// Renove named actor from the scene
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (*it == actor)
		{
			// Do not delete managed actors as the pooled object manager will be handling this
			if (!(*it)->isManaged())
				delete *it;
			else
			{
				// Instead we disable them and mark them as unused so they can be reallocated
				(*it)->setUsed(false);
				(*it)->setActive(false);
			}
			Actors.erase(it);
			break;
		}
	}
}

void CIwGameScene::removeActor(unsigned int name_hash)
{
	// Renove named actor from the scene
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (name_hash == (*it)->getNameHash())
		{
			// Do not delete managed actors as the pooled object manager will be handling this
			if (!(*it)->isManaged())
				delete *it;
			else
			{
				// Instead we disable them and mark them as unused so they can be reallocated
				(*it)->setUsed(false);
				(*it)->setActive(false);
			}
			Actors.erase(it);
			break;
		}
	}
}

CIwGameActor* CIwGameScene::findActor(const char* name)
{
	return findActor(CIwGameString::CalculateHash(name));
}

CIwGameActor* CIwGameScene::findActor(unsigned int name_hash)
{
	// Find and eturn named actor
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (name_hash == (*it)->getNameHash())
			return *it;
	}

	return NULL;
}

CIwGameActor* CIwGameScene::findActor(int type)
{
	// Find and return first actor of specified type
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (type == (*it)->getType())
			return *it;
	}
	return NULL;
}

void CIwGameScene::clearActors()
{
	// Remove and delete all actors from the scsne (only delete if actors are not managed)
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (!(*it)->isManaged())
			delete *it;
	}
	Actors.clear();
}

bool CIwGameScene::addCollideable(CIwGameActor* actor)
{
	if (NextFreeCollidable < MaxCollidables)
		Collidables[NextFreeCollidable++] = actor;
	else
		return false;

	return true;
}

void CIwGameScene::Update(float dt)
{
	if (!IsActive)
		return;

	if (Camera != NULL)
	{
		// I need to point out here that we use the concept of dirty transforms to cut down processing when  the camera is not moving
		// If the cameras transform is not dirty (the camera has not moved) then there is no need to force the sprite manager to make 
		// all of its child sprites to rebuild their transforms
		if (Camera->isTransformDirty())
		{
			// Update camera transform
			Camera->Update();

			// Combine the virtual canvas and camera transforms
			Transform = Camera->getTransform();
			Transform.PostMult(VirtualTransform);

			// Transfer scene transform to sprite manager
			if (SpriteManager != NULL)
			{
				SpriteManager->setTransform(Transform);
			}
		}
	}

	// Process the actors
	CIwList<CIwGameActor*> removals;
	for (_Iterator it = begin(); it != end(); ++it)
	{
		if (!(*it)->Update(dt))
			removals.push_back(*it);
	}

	// Resolve actor collisions
	for (_Iterator it = begin(); it != end(); ++it)
	{
		if ((*it)->isCollidable())
			(*it)->ResolveCollisions();
	}

	// Remove any deleted actors
	for (_Iterator it = removals.begin(); it != removals.end(); ++it)
		removeActor(*it);

	// Update the animation timeline
	if (Timeline != NULL)
 		Timeline->Update(dt);

	// Reset collidables list
	NextFreeCollidable = 0;
}

void CIwGameScene::Draw()
{
	if (!IsVisible)
		return;

	// Transfor actor visual data to sprite manager sprites
	for (_Iterator it = begin(); it != end(); ++it)
		(*it)->UpdateVisual();

	if (ClippingArea.w != 0)
	{
		// Set clipping rectangle (clipping coords are specified in virtual canvas coordinates so we need to transform them to the screen)
		CIwSVec2 cliprect[2];
		cliprect[0].x = ClippingArea.x;
		cliprect[1].x = ClippingArea.x + ClippingArea.w;
		cliprect[0].y = ClippingArea.y;
		cliprect[1].y = ClippingArea.y + ClippingArea.h;
		for (int t = 0; t < 2; t++)
			cliprect[t] = VirtualTransform.TransformVec(cliprect[t]);
		IwGxClearScissorScreenSpace();
		IwGxSetScissorScreenSpace(cliprect[0].x, cliprect[0].y, cliprect[1].x - cliprect[0].x, cliprect[1].y - cliprect[0].y);
	}

	// Draw the sprite manager sprites
	SpriteManager->Draw();

}

CIwFVec2 CIwGameScene::ScreenToVirtual(float screen_x, float screen_y)
{
	screen_x -= ScreenSize.x / 2;
	screen_y -= ScreenSize.y / 2;
	return CIwFVec2((screen_x * IW_GEOM_ONE) / Transform.m[0][0], (screen_y * IW_GEOM_ONE) / Transform.m[1][1]);
}

CIwVec2 CIwGameScene::VirtualToScreen(float virtual_x, float virtual_y)
{
	return CIwVec2((int)((virtual_x * Transform.m[0][0]) / IW_GEOM_ONE + ScreenSize.x / 2), (int)((virtual_y * Transform.m[1][1]) / IW_GEOM_ONE + ScreenSize.y / 2));
}

bool CIwGameScene::LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node)
{
	// Process scene attributes
	int				max_colliders = 128;
	int				max_layers = 10;
	CIwFVec2		canvas_size((float)IwGxGetScreenWidth(), (float)IwGxGetScreenHeight());
	bool			fix_aspect = true;
	bool			lock_width = false;
	bool			current = false;
	CIwGameString	timeline_name;
	CIwColour		colour;

	colour.Set(0xffffffff);

	for (CIwGameXmlNode::_AttribIterator it = node->attribs_begin(); it != node->attribs_end(); it++)
	{
		unsigned int name_hash = (*it)->getName().getHash();

		if (name_hash == CIwGameXomlNames::Name_Hash)
		{
			setName((*it)->GetValue().c_str());
		}
		else
		if (name_hash == CIwGameXomlNames::CanvasSize_Hash)
		{
			if (!(*it)->GetValueAsPoint(canvas_size))
				CIwGameError::LogError("Warning: XOML - Invalid value for Scene::CanvasSize");
		}
		else
		if (name_hash == CIwGameXomlNames::FixAspect_Hash)
		{
			fix_aspect = (*it)->GetValueAsBool();
		}
		else
		if (name_hash == CIwGameXomlNames::LockWidth_Hash)
		{
			lock_width = (*it)->GetValueAsBool();
		}
		else
		if (name_hash == CIwGameXomlNames::Extents_Hash)
		{
			CIwRect extents;
			if (!(*it)->GetValueAsRect(extents))
				CIwGameError::LogError("Warning: XOML - Invalid value for Scene::Extents");
			else
				setExtents(extents.x, extents.y, extents.w, extents.h);
		}
		else
		if (name_hash == CIwGameXomlNames::AllowSuspend_Hash)
		{
			setAllowSuspend((*it)->GetValueAsBool());
		}
		else
		if (name_hash == CIwGameXomlNames::Clipping_Hash)
		{
			CIwRect clipping;
			if (!(*it)->GetValueAsRect(clipping))
				CIwGameError::LogError("Warning: XOML - Invalid value for Scene::Clipping");
			else
				setClippingArea(clipping.x, clipping.y, clipping.w, clipping.h);
		}
		else
		if (name_hash == CIwGameXomlNames::Active_Hash)
		{
			setActive((*it)->GetValueAsBool());
		}
		else
		if (name_hash == CIwGameXomlNames::Visible_Hash)
		{
			setVisible((*it)->GetValueAsBool());
		}
		else
		if (name_hash == CIwGameXomlNames::Layers_Hash)
		{
			max_layers = ((*it)->GetValueAsInt());
		}
		else
		if (name_hash == CIwGameXomlNames::Colliders_Hash)
		{
			max_colliders = ((*it)->GetValueAsInt());
		}
		else
		if (name_hash == CIwGameXomlNames::Current_Hash)
		{
			current = ((*it)->GetValueAsBool());
		}
		else
		if (name_hash == CIwGameXomlNames::Timeline_Hash)
		{
			timeline_name = (*it)->GetValue();
		}
		else
		if (name_hash == CIwGameXomlNames::Colour_Hash)
		{
			if (!(*it)->GetValueAsColour(colour))
			{
				CIwGameError::LogError("Warning: XOML - Invalid value for Scene::Colour");
			}
		}
	}

	// Initialise the scene
	if (load_children)
		Init(max_colliders, max_layers);

	// Set the colour
	Colour = colour;

	// Set the virtual tansform
	setVirtualTransform((int)canvas_size.x, (int)canvas_size.y, 0, fix_aspect, lock_width);

	if (load_children)
	{
		// Process inner tags
		if (!IW_GAME_XOML->Process(this, node))
			return false;
	}

	// Check that we were created from a CIwGame derived class
	if (parent->getClassType() != CIwGameXomlNames::Game_Hash)
	{
		CIwGameError::LogError("Error: XOML - Scene needs to be declared inside a game type class");
		return false;
	}

	// Search the scenes resources for the specified timeline
	if (!timeline_name.IsEmpty())
	{
		CIwGameAnimTimeline* timeline = getTimelineManager()->findTimeline(timeline_name.c_str());
		if (timeline == NULL)
		{
			CIwGameError::LogError("Error: XOML - Could not find named timeline in the scene or in the global resources - ", timeline_name.c_str());
			return false;
		}
		setTimeline(timeline);
		timeline->setTargetElement(this);
	}

	// Add the scene to the game
	((CIwGame*)parent)->addScene(this);

	if (current)
	{
		// If scene is marked as current then switcn to it
		((CIwGame*)parent)->changeScene(this);
	}

	return true;
}

bool CIwGameScene::UpdateFromAnimation(CIwGameAnimInstance *animation)
{
	unsigned int element_name = animation->getTargetPropertyHash();

	if (Camera != NULL)
	{
		if (element_name == CIwGameXomlNames::Position_Hash)
		{
			CIwGameAnimFrameVec2* frame = (CIwGameAnimFrameVec2*)animation->getCurrentData();
			Camera->setPosition(frame->data.x, frame->data.y);
			return true;
		}
		else
		if (element_name == CIwGameXomlNames::Angle_Hash)
		{
			CIwGameAnimFrameFloat* frame = (CIwGameAnimFrameFloat*)animation->getCurrentData();
			Camera->setAngle(frame->data);
			return true;
		}
		else
		if (element_name == CIwGameXomlNames::Scale_Hash)
		{
			CIwGameAnimFrameFloat* frame = (CIwGameAnimFrameFloat*)animation->getCurrentData();
			Camera->setScale(frame->data);
			return true;
		}
	}
	if (element_name == CIwGameXomlNames::Colour_Hash)
	{
		CIwGameAnimFrameVec4* frame = (CIwGameAnimFrameVec4*)animation->getCurrentData();
		Colour.r = (uint8)frame->data.x;
		Colour.g = (uint8)frame->data.y;
		Colour.b = (uint8)frame->data.z;
		Colour.a = (uint8)frame->data.w;
		return true;
	}
	else
	if (element_name == CIwGameXomlNames::Clipping_Hash)
	{
		CIwGameAnimFrameRect* frame = (CIwGameAnimFrameRect*)animation->getCurrentData();
		ClippingArea.x = (uint8)frame->data.x;
		ClippingArea.y = (uint8)frame->data.y;
		ClippingArea.w = (uint8)frame->data.w;
		ClippingArea.h = (uint8)frame->data.h;
		return true;
	}
	else
	if (element_name == CIwGameXomlNames::Visible_Hash)
	{
		CIwGameAnimFrameBool* frame = (CIwGameAnimFrameBool*)animation->getCurrentData();
		IsVisible = frame->data;
		return true;
	}
	else
	if (element_name == CIwGameXomlNames::Timeline_Hash)
	{
		CIwGameAnimFrameString* frame = (CIwGameAnimFrameString*)animation->getCurrentData();
		if (!frame->data.IsEmpty())
		{
			CIwGameAnimTimeline* timeline = getTimelineManager()->findTimeline(frame->data.c_str());
			if (timeline != NULL && timeline != Timeline)
			{
				setTimeline(timeline);
				timeline->setTargetElement(this);
			}
		}
		return true;
	}

	return false;
}

